"""
Defines models for power systems components, including
:class:`~powersystems.PowerSystem`, :class:`~powersystems.Bus`,
:class:`~powersystems.Load` and  :class:`~powersystems.Line`.
:class:`~powersystems.Generator` components can be found
in the :module:`~generators`. Each of these objects inherits an
optimization framework from :class:`~optimization.OptimizationObject`.
"""

import logging

from commonscripts import (update_attributes, getattrL, flatten, 
    quiet, set_trace)
from config import user_config

from optimization import (value, OptimizationObject,
                          OptimizationProblem, OptimizationError)
import stochastic


from coopr import pyomo
import numpy as np
import pandas as pd


class Load(OptimizationObject):
    """
    Describes a power system load (demand).
    Currently only real power is considered.
    For OPF problems, the name of the bus can.
    For UC problems, schedules (pandas.Series objects) are used.
    By setting `sheddingallowed`, the amount of power can become a variable,
        (bounded to be at most the scheduled amount).
    """
    def __init__(self,
                 name='', index=None, bus=None, schedule=None,
                 sheddingallowed=True,
                 cost_shedding=None,
                 ):
        update_attributes(self, locals())  # load in inputs
        if cost_shedding is None:
            self.cost_shedding = user_config.cost_load_shedding
        self.init_optimization()
        self.shedding_mode = False

    def power(self, time, scenario=None, evaluate=False):
        if self.shedding_mode:
            power = self.get_variable('power', time, 
                scenario=scenario, indexed=True)
            if evaluate:
                power = value(power)
            return power
        else:
            return self.get_scheduled_output(time)

    def shed(self, time, scenario=None, evaluate=False):
        return self.get_scheduled_output(time) - self.power(time, scenario, evaluate)

    def cost(self, time, scenario=None):
        return self.cost_shedding * self.shed(time, scenario)

    def cost_first_stage(self, times):
        return 0

    def cost_second_stage(self, times):
        return sum(self.cost(time) for time in times)

    def create_variables(self, times):
        if self.shedding_mode:
            self.add_variable('power', index=times.set, low=0)

    def create_constraints(self, times):
        if self.shedding_mode:
            for time in times:
                self.add_constraint('max_load_power', time,
                                    self.power(time) <= self.get_scheduled_output(time))

    def create_objective(self, times):
        return sum([self.cost(time) for time in times])

    def __str__(self):
        return 'd{ind}'.format(ind=self.index)

    def __int__(self):
        return self.index

    def get_scheduled_output(self, time):
        return float(self.schedule.ix[time])


class Line(OptimizationObject):
    """
    A tranmission line. Currently the model
    only considers real power flow under normal conditions.
    """
    def __init__(self, name='', index=None, frombus=None, tobus=None,
                 reactance=0.05, pmin=None, pmax=9999, **kwargs):
        update_attributes(self, locals())  # load in inputs
        if self.pmin is None:
            self.pmin = -1 * self.pmax   # default is -1*pmax
        self.init_optimization()

    def power(self, time):
        return self.get_variable('power', time, indexed=True)

    def price(self, time):
        '''congestion price on line'''
        return self.get_dual('line flow', time)

    def create_variables(self, times):
        self.add_variable('power', index=times.set)

    def create_constraints(self, times, buses):
        '''create the constraints for a line over all times'''
        busNames = getattrL(buses, 'name')
        iFrom, iTo = busNames.index(self.frombus), busNames.index(self.tobus)
        for t in times:
            line_flow_ij = self.power(t) == \
                1 / self.reactance * (buses[iFrom].angle(t) -
                                      buses[iTo].angle(t))
            self.add_constraint('line flow', t, line_flow_ij)
            self.add_constraint(
                'line limit high', t, self.power(t) <= self.pmax)
            self.add_constraint(
                'line limit low', t, self.pmin <= self.power(t))
        return

    def __str__(self):
        return 'k{ind}'.format(ind=self.index)

    def __int__(self):
        return self.index


class Bus(OptimizationObject):
    """
    A transmission bus bus (usually a substation where one or more
    tranmission lines start/end).

    :param isSwing: flag if the bus is the swing bus
      (sets the reference angle for the system)
    """
    def __init__(self, name=None, index=None, isSwing=False):
        update_attributes(self, locals())  # load in inputs
        self.generators, self.loads = [], []
        self.init_optimization()

    def angle(self, time):
        return self.get_variable('angle', time, indexed=True)

    def price(self, time):
        return self.get_dual('power balance', time)

    def Pgen(self, t, evaluate=False):
        if evaluate:
            return sum(value(gen.power(t)) for gen in self.generators)
        else:
            return sum(gen.power(t) for gen in self.generators)

    def Pload(self, t, evaluate=False):
        if evaluate:
            return sum(value(ld.power(t)) for ld in self.loads)
        else:
            return sum(ld.power(t) for ld in self.loads)

    def power_balance(self, t, Bmatrix, allBuses):
        if len(allBuses) == 1:
            lineFlowsFromBus = 0
        else:
            lineFlowsFromBus = sum([Bmatrix[self.index][otherBus.index] * otherBus.angle(t) for otherBus in allBuses])  # P_{ij}=sum_{i} B_{ij}*theta_j ???
        return sum([-lineFlowsFromBus, -self.Pload(t), self.Pgen(t)])

    def create_variables(self, times):
        self.add_children(self.generators, 'generators')
        self.add_children(self.loads, 'loads')
        logging.debug('added bus {} components - generators and loads'.format(
            self.name))
#        if len(self.generators)<50:
        for gen in self.generators:
            gen.create_variables(times)
#        else:
#            for gen in self.generators:
#                threading.Thread(target=_call_generator_create_variables,args=(gen,times)).start()
#            else:
#                for th in threading.enumerate():
#                    if th is threading.current_thread(): continue
#                    else: th.join()

        logging.debug('created generator variables')
        for load in self.loads:
            load.create_variables(times)
        logging.debug('created load variables')
        self.add_variable('angle', index=times.set)
        logging.debug('created bus variables ... returning')
        return

    def create_objective(self, times):
        return self.cost_first_stage(times) + self.cost_second_stage(times)

    def cost_first_stage(self, times):
        return sum(gen.cost_first_stage(times) for gen in self.generators) + \
            sum(load.cost_first_stage(times) for load in self.loads)

    def cost_second_stage(self, times):
        return sum(gen.cost_second_stage(times) for gen in self.generators) + \
            sum(load.cost_second_stage(times) for load in self.loads)

    def create_constraints(self, times, Bmatrix, buses, include_children=True):
        if include_children:
            for gen in self.generators:
                gen.create_constraints(times)
            for load in self.loads:
                load.create_constraints(times)
        nBus = len(buses)
        for time in times:
            self.add_constraint('power balance', time, self.power_balance(
                time, Bmatrix, buses) == 0)  # power balance must be zero
            if nBus > 1 and self.isSwing:
                self.add_constraint('swing bus', time, self.angle(
                    time) == 0)  # swing bus has angle=0
        return
    # def clear_constraints(self):
    #     self.constraints={}
    #     for gen in self.generators: gen.clear_constraints()
    #     for load in self.loads: load.clear_constraints()

    def __str__(self):
        return 'i{ind}'.format(ind=self.index)


class PowerSystem(OptimizationProblem):
    '''
    Power systems object which is the container for all other components.

    :param generators: list of :class:`~powersystem.Generator` objects
    :param loads: list of :class:`~powersystem.Load` objects
    :param lines: list of :class:`~powersystem.Line` objects

    Other settings are inherited from `user_config`.
    '''
    def __init__(self, generators, loads, lines=None):
        # load in inputs
        update_attributes(self, locals(),
                          exclude=['generators', 'loads', 'lines'])
        self.reserve_fixed = user_config.reserve_fixed
        self.reserve_load_fraction = user_config.reserve_load_fraction
        self.reserve_required = (self.reserve_fixed > 0) or \
            (self.reserve_load_fraction > 0.0)
        
        if lines is None:
            lines = []

        buses = self.make_buses_list(loads, generators)
        self.create_admittance_matrix(buses, lines)
        self.init_optimization()

        self.add_children(buses, 'buses')
        self.add_children(lines, 'lines')

        self.is_stochastic = len(
            filter(lambda gen: gen.is_stochastic, generators)) > 0
        self.shedding_mode = False

    def make_buses_list(self, loads, generators):
        """
        Create list of :class:`powersystems.Bus` objects
        from the load and generator bus names. Otherwise
        (as in ED,UC) create just one (system)
        :class:`powersystems.Bus` instance.

        :param loads: a list of :class:`powersystems.Load` objects
        :param generators: a list of :class:`powersystems.Generator` objects
        :returns: a list of :class:`powersystems.Bus` objects
        """
        busNameL = []
        busNameL.extend(getattrL(generators, 'bus'))
        busNameL.extend(getattrL(loads, 'bus'))
        busNameL = pd.Series(pd.unique(busNameL)).dropna().tolist()

        if len(busNameL) == 0:
            busNameL = [None]

        buses = []
        swingHasBeenSet = False

        for b, busNm in enumerate(busNameL):
            newBus = Bus(name=busNm, index=b)
            for gen in generators:
                if gen.bus == newBus.name:
                    newBus.generators.append(gen)
                if not swingHasBeenSet:
                    newBus.isSwing = swingHasBeenSet = True
            for ld in loads:
                if ld.bus == newBus.name:
                    newBus.loads.append(ld)
            buses.append(newBus)
        return buses

    def create_admittance_matrix(self, buses, lines):
        """
        Creates the admittance matrix (B),
        with elements = total admittance of line from bus i to j.
        Used in calculating the power balance for OPF problems.

        :param buses: list of :class:`~powersystems.Line` objects
        :param lines: list of :class:`~powersystems.Bus` objects
        """
        nB = len(buses)
        self.Bmatrix = np.zeros((nB, nB))
        namesL = [bus.name for bus in buses]
        for line in lines:
            busFrom = buses[namesL.index(line.frombus)]
            busTo = buses[namesL.index(line.tobus)]
            self.Bmatrix[busFrom.index, busTo.index] += -1 / line.reactance
            self.Bmatrix[busTo.index, busFrom.index] += -1 / line.reactance
        for i in range(0, nB):
            self.Bmatrix[i, i] = -1 * sum(self.Bmatrix[i, :])

    def loads(self):
        return flatten(bus.loads for bus in self.buses)

    def generators(self):
        return flatten(bus.generators for bus in self.buses)

    def create_variables(self, times):
        self.add_variable('cost_first_stage')
        self.add_variable('cost_second_stage')
        self.add_set('times', times._set, ordered=True)
        times.set = self._model.times
        for bus in self.buses:
            bus.create_variables(times)
        for line in self.lines:
            line.create_variables(times)
        logging.debug('... created power system vars... returning')

    def cost_first_stage(self, scenario=None):
        return self.get_component('cost_first_stage', scenario=scenario)

    def cost_second_stage(self, scenario=None):
        return self.get_component('cost_second_stage', scenario=scenario)

    def create_objective(self, times):
        self.add_objective(self.cost_first_stage() + self.cost_second_stage())

    def create_constraints(self, times, include_children=True):
        if include_children:
            if user_config.duals: 
                self.add_suffix('dual')
            for bus in self.buses:
                bus.create_constraints(times, self.Bmatrix, self.buses)
            for line in self.lines:
                line.create_constraints(times, self.buses)
        
        # system reserve constraint
        self._has_reserve = not self.shedding_mode and \
            (self.reserve_fixed > 0 or self.reserve_load_fraction > 0)
        if self._has_reserve:
            for time in times:
                required_generation_availability = self.reserve_fixed + (1.0 + self.reserve_load_fraction) * sum(load.power(time) for load in self.loads())
                generation_availability = sum(
                    gen.power_available(time) for gen in self.generators())
                self.add_constraint('reserve', generation_availability >= required_generation_availability, time=time)

        self.add_constraint('system_cost_first_stage', 
            self.cost_first_stage() == \
            sum(bus.cost_first_stage(times) for bus in self.buses))
        self.add_constraint('system_cost_second_stage', 
            self.cost_second_stage() == \
            sum(bus.cost_second_stage(times) for bus in self.buses))



    def iden(self, time=None):
        name = 'system'
        if time is not None:
            name += '_' + str(time)
        return name

    def total_scheduled_load(self):
        return sum([load.schedule for load in self.loads()])

    def total_scheduled_generation(self):
        return sum(gen.schedule for gen in self.generators() if not gen.is_controllable)

    def get_generators_controllable(self):
        return filter(lambda gen: gen.is_controllable, self.generators())

    def get_generators_noncontrollable(self):
        return filter(lambda gen: not gen.is_controllable, self.generators())

    def get_generators_without_scenarios(self):
        return filter(lambda gen: getattr(gen, 'is_stochastic', False) == False, self.generators())

    def get_generator_with_scenarios(self):
        gens = filter(lambda gen: getattr(gen, 'is_stochastic',
                      False), self.generators())
        if len(gens) > 1:
            raise NotImplementedError(
                'Dont handle the case of multiple stochastic generators')
        elif len(gens) == 0:
            return []
        else:
            return gens[0]

    def get_generator_with_observed(self):
        return filter(lambda gen: getattr(gen, 'observed_values', None) is not None, self.generators())[0]

    def get_finalconditions(self, sln):
        times = sln.times

        tEnd = times.last_non_overlap()  # like 2011-01-01 23:00:00
        tEndstr = times.non_overlap().last()  # like t99

        status = sln.generators_status

        for gen in self.generators():
            g = str(gen)
            stat = status[g]
            if sln.is_stochastic:
                gen.finalstatus = dict(
                    power=sln.generators_power[g][tEnd],
                    status=sln.generators_status[g][tEnd],
                    hoursinstatus=gen.gethrsinstatus(times.non_overlap(), stat)
                )
            else:
                gen.finalstatus = gen.getstatus(tEndstr,
                                                times.non_overlap(), stat)
        return

    def set_initialconditions(self, initTime):
        for gen in self.generators():
            finalstatus = getattr(gen, 'finalstatus', {})
            if finalstatus:
                gen.set_initial_condition(time=initTime, **finalstatus)
                del gen.finalstatus
        return

    def solve_problem(self, times):
        try:
            instance = self.solve()
                            
        except OptimizationError:
            # re-do stage, with load shedding allowed
            logging.critical('stage infeasible, re-run with shedding.')
            self.allow_shedding(times)
            try:
                instance = self.solve()
            except OptimizationError:
                scheduled, committed = self.debug_infeasible(times)
                raise OptimizationError('failed to solve with shedding.')
        return instance

    def resolve_stochastic_with_observed(self, instance, sln):
        s = sln.scenarios[0]
        self._model = instance.active_components(pyomo.Block)[s]
        self.is_stochastic = False
        self.stochastic_formulation = False

        self._resolve_problem(sln)

        # re-store the generator outputs and costs
        sln._resolved = True
        sln._get_outputs(resolve=True)
        sln._get_costs(resolve=True)

        self.is_stochastic = True
        self.disallow_shedding()
        return

    def resolve_determinisitc_with_observed(self, sln):
        # store the useful expected value solution information
        sln.expected_status = sln.generators_status.copy()
        sln.expected_power = sln.generators_power.copy()
        sln.expected_fuelcost = sln.fuelcost.copy()
        sln.expected_totalcost = sln.totalcost_generation.copy()
        sln.expected_load_shed = float(sln.load_shed)

        # resolve the problem
        self._resolve_problem(sln)

        # re-calc the generator outputs and costs
        sln._resolved = True
        sln._get_outputs()
        sln._get_costs()

        sln.observed_fuelcost = sln.fuelcost
        sln.observed_totalcost = sln.totalcost_generation
        self.disallow_shedding()
        return

    def _set_load_shedding(self, to_mode):
        '''set system mode for load shedding'''
        for load in filter(lambda ld:
                           ld.sheddingallowed, self.loads()):
            load.shedding_mode = to_mode

    def _set_gen_shedding(self, to_mode):
        for gen in filter(lambda g:
            not g.is_controllable and g.sheddingallowed, self.generators()):
            gen.shedding_mode = to_mode

    def allow_shedding(self, times, resolve=False):
        self.shedding_mode = True
        self._set_load_shedding(True)
        
        if not user_config.economic_wind_shed:
            logging.debug('allowing non-controllable generation shedding')
            self._set_gen_shedding(True)

        const_times = times.non_overlap() if resolve else times

        # make load power into a variable instead of a param
        for load in self.loads():
            try: 
                load.create_variables(times)  # need all for the .set attrib
                load.create_constraints(const_times)
            except RuntimeError:
                # load already has a power variable and shedding constraint
                pass
        
        if not user_config.economic_wind_shed:
            for gen in filter(lambda gen: gen.shedding_mode,
                self.get_generators_noncontrollable()):
                # create only the power_used var, don't reset the power param
                gen.create_variables_shedding(times)
                gen.create_constraints(const_times)

        # recalc the power balance constraint
        for bus in self.buses:
            for time in times:
                bus._remove_component('power balance', time)
            bus.create_constraints(const_times,
                self.Bmatrix, self.buses, include_children=False)

        # reset objective
        self.reset_objective()
        self.create_objective(const_times)
        # re-create system cost constraints
        self._remove_component('system_cost_first_stage')
        self._remove_component('system_cost_second_stage')
        if self._has_reserve:
            self._remove_component('reserve')
        self.create_constraints(const_times, include_children=False)

        # recreating all constraints would be simpler
        # but would take a bit longer
        # self.create_constraints(const_times, include_children=True)        
        if self.is_stochastic:
            # need to recreate the scenario tree variable links 
            stochastic.define_stage_variables(self, times)
            # and the stochastic instance            
            stochastic.create_problem_with_scenarios(self, times)
            

    def disallow_shedding(self):
        # change shedding allowed flags for the next stage
        self.shedding_mode = False
        self._set_load_shedding(False)
        if user_config.economic_wind_shed is False:
            self._set_gen_shedding(False)

    def _resolve_problem(self, sln):
        times = sln.times_non_overlap
        self._remove_component('times')
        self.add_set('times', times._set, ordered=True)
        times.set = self._model.times
        
        # reset the constraints
        self._remove_all_constraints()
        # dont create reserve constraints
        self.reserve_fixed = 0
        self.reserve_load_fraction = 0

        # set wind to observed power
        gen = self.get_generator_with_observed()
        gen.set_power_to_observed(times)

        # reset objective to only the non-overlap times
        self.reset_objective()
        self.create_objective(times)

        # recreate constraints only for the non-overlap times
        self.create_constraints(times)

        # fix statuses for all units
        self.fix_binary_variables()

        # store original problem solve time
        self.full_sln_time = self.solution_time
        full_mipgap = self.mipgap

        logging.info('resolving with observed values')
        try:
            self.solve()
        except OptimizationError:
            faststarts = map(lambda gen: str(gen), filter(lambda gen: gen.faststart, self.generators()))
            # at least one faststarting unit must be available (off)
            if user_config.faststart_resolve and \
                (sln.expected_status[faststarts] == 0).any().any():
                self._resolve_with_faststarts(sln)
            else:
                # just shed the un-meetable load and calculate cost later
                self.allow_shedding(sln.times, resolve=True)
                try:
                    self.solve()
                except OptimizationError:
                    scheduled, committed = self.debug_infeasible(
                        sln.times, resolve_sln=sln)
                    raise

        self.resolve_solution_time = float(self.solution_time)
        self.solution_time = float(self.full_sln_time)

        if self.mipgap:
            self.mipgap_resolve = float(self.mipgap)
        if full_mipgap:
            self.mipgap = float(full_mipgap)

        logging.info('resolved instance with observed values (in {}s)'.format(
            self.resolve_solution_time))

    def _resolve_with_faststarts(self, sln):
        '''allow faststart units to be started up to meet the load'''
        self._unfix_variables()
        self._fix_non_faststarts(sln.times)
        logging.warning('allowing fast-starting units')
        
        try:
            self.solve()
        except OptimizationError:
            self._unfix_variables()
            self._fix_non_faststarts(sln.times, fix_power=False)
            logging.warning(
                'allowing non fast-starters to change power output')
            try:
                self.solve()
            except OptimizationError:
                logging.warning('allowing load shedding')
                self.allow_shedding(sln.times, resolve=True)
                try:
                    self.solve()
                except OptimizationError:
                    scheduled, committed = self.debug_infeasible(
                        sln.times, resolve_sln=sln)
                    raise

    def _fix_non_faststarts(self, times, fix_power=True):
        '''
        fix non-faststart units - both power and status
        (unless this is infeasible, then only fix status)
        the idea is that fast-starts should be contributing power
        only for system security, not economics
        '''
        names = []
        for gen in filter(lambda gen:
                         (not gen.faststart) and gen.is_controllable, self.generators()):
            names.append(gen.status().name)
            if fix_power:
                names.append(gen.power().name)
        self._fix_variables(names)

    def debug_infeasible(self, times, resolve_sln=None):
        generators = self.generators()
        if resolve_sln:
            windgen = self.get_generator_with_observed()

            scheduled = pd.DataFrame({
                'expected_power': resolve_sln.generators_power.sum(axis=1).values,
                'expected_wind': windgen.schedule.ix[times.non_overlap()],
                'observed_wind': windgen.observed_values.ix[times.non_overlap()],
            })
            scheduled['net_required'] = \
                scheduled.expected_wind - scheduled.observed_wind
        else:
            scheduled = pd.DataFrame({
                'load': self.total_scheduled_load().ix[times.strings.values]})

            if self.is_stochastic:
                gen = self.get_generator_with_scenarios()
                scenarios = gen.scenario_values[times.Start].drop('probability', axis=1).T
                scenarios.index = scheduled.index
                
                scheduled['net_load'] = scheduled['load'] - sum(
                    map(lambda gen: gen.schedule, 
                        filter(lambda gen: not gen.is_stochastic, 
                            self.get_generators_noncontrollable())))
                
                gen_required = (-1 * scenarios).add(scheduled.net_load, axis=0)
                
                print('generation required')
                print(gen_required)
                print(gen_required.describe())
                
            else:                    
                if any([hasattr(gen, 'schedule') for gen in self.generators()]):
                    scheduled['generation'] = self.total_scheduled_generation().ix[times.strings.values]
                else:
                    scheduled['generation'] = 0
                    
                scheduled['net_required'] = scheduled['load'] - \
                    scheduled.generation
                    
        print 'total scheduled\n', scheduled

        if resolve_sln:
            committed = pd.DataFrame(dict(
                Pmin=[gen.pmin for gen in generators],
                Pmax=[gen.pmax for gen in generators],
                rampratemin=[getattr(gen, 'rampratemin',
                                     None) for gen in generators],
                rampratemax=[getattr(gen, 'rampratemax',
                                     None) for gen in generators],
            )).T
            print 'generator limits\n', committed
        else:
            gens = filter(lambda gen:
                          gen.is_controllable and gen.initial_status == 1,
                          self.generators())
            committed = pd.Series(dict(
                Pmin=sum(gen.pmin for gen in gens),
                Pmax=sum(gen.pmax for gen in gens),
                rampratemin=pd.Series(gen.rampratemin for gen in gens).sum(),
                rampratemax=pd.Series(gen.rampratemax for gen in gens).sum(),
            ))
            print 'total committed\n', committed

        if resolve_sln:
            print 'expected status'
            if len(resolve_sln.generators_status.columns) < 5: 
                print(resolve_sln.generators_status)
            else:
                print(resolve_sln.generators_status.sum(axis=1))
            ep = resolve_sln.generators_power
            ep['net_required'] = scheduled.net_required.values
            print('expected power')
            if len(ep.columns) < 5:
                print(ep)
            else: 
                print(ep.sum(axis=1))
        else:
            print 'initial_status\n'
            print pd.Series([gen.initial_status for gen in self.generators()])
        
        return scheduled, committed



"""
A module to workaround a memory leak in coopr by saving the results
of each day of a rolling unit commitment to disk (in HDF format)
and reloading them to run the next day as a memory independent subprocess.
"""

import os
import logging
import sys
import pandas as pd
from pandas import Series, DataFrame
from commonscripts import gen_time_dataframe, debug_frame_unequal, set_trace
from config import user_config

from results import _correct_status
from schedule import TimeIndex
from get_data import parse_standalone
import pkg_resources


try:
    import tables as tb
except ImportError:
    logging.debug('could not load pytables - cannot use standalone mode.')


def wipe_storage():
    try:
        os.remove(user_config.store_filename)
        logging.debug('wipe storage')
    except OSError:
        pass


def get_storage():
    return pd.HDFStore(user_config.store_filename)


def store_times(tstage, storage=None):
    if storage is None:
        storage = get_storage()
    storage['times'] = tstage.strings
    return storage


def init_store(power_system, times, data):
    '''create the store before the first stage'''
    wipe_storage()
    storage = get_storage()
    generators = power_system.generators()

    # store the problem info read from the spreadsheets
    for key, df in data.iteritems():
        if key != 'scenario_values': 
            for k, v in (df.dtypes == object).iterkv():
                if v: df[k] = df[k].fillna('')
        
        key = 'data_' + key
        storage[key] = df

    stages = len(times)  # stage_times
    est_Nt = len(data['timeseries'])
    t = [times[0].initialTime]

    # store first stage initial state
    storage.append('power', 
        gen_time_dataframe(generators, t,
            values=[[gen.initial_power for gen in generators]]),
        expectedrows=est_Nt)
        
    storage.append('status',
        gen_time_dataframe(generators, t,
            values=[[gen.initial_status for gen in generators]]),
        expectedrows=est_Nt)
        
    storage['hrsinstatus'] = gen_time_dataframe(generators, t,
        values=[[gen.initial_status_hours for gen in generators]])

    # setup empty containers for variables
    storage['load_shed'] = Series()
    storage['gen_shed'] = Series()
    storage['expected_cost'] = DataFrame()
    storage['observed_cost'] = DataFrame()
    storage['expected_status'] = DataFrame()
    storage['expected_power'] = DataFrame()
    storage['expected_fuelcost'] = DataFrame()
    storage['observed_fuelcost'] = DataFrame()

    # setup one-per-stage results
    storage['solve_time'] = Series(index=range(stages))
    storage['mipgap'] = Series(index=range(stages))

    # store configuration
    storage['configuration'] = Series(user_config)

    storage['version'] = Series({
        'minpower': pkg_resources.get_distribution('minpower').version,
        # try storing version number of the current working directory
        'problem': _get_problem_version(),
    })

    return storage


def _get_problem_version():
    version = ''
    possible_init = os.path.join(
        os.path.abspath(user_config.directory), '__init__.py')
    if os.path.exists(possible_init):
        vstr = open(possible_init, 'r').read()
        version = vstr.split('"')[1].rstrip('"')
    return version


def store_state(power_system, times, sln=None):
    storage = get_storage()
    generators = power_system.generators()

    stg = sln.stage_number
    table_append(storage, 'power', sln.generators_power)
    table_append(storage, 'status', sln.generators_status)
    table_append(storage, 'load_shed', sln.load_shed_timeseries)
    table_append(storage, 'gen_shed', sln.gen_shed_timeseries)

    tEnd = times.last_non_overlap()
    storage['hrsinstatus'] = gen_time_dataframe(generators, [tEnd],
                                                values=[
                                                [gen.finalstatus['hoursinstatus'] for gen in generators]
                                                ])

    _add_tbl_val(storage, 'solve_time', stg, sln.solve_time)
    _add_tbl_val(storage, 'mipgap', stg, sln.mipgap)

    if sln._resolved:
        table_append(storage, 'observed_cost', sln.observed_totalcost)
        table_append(storage, 'observed_fuelcost', sln.observed_fuelcost)
        table_append(storage, 'expected_cost', sln.expected_totalcost)
        table_append(storage, 'expected_fuelcost', sln.expected_fuelcost)
        table_append(storage, 'expected_power', sln.expected_power)
        table_append(storage, 'expected_status', sln.expected_status)
    else:
        table_append(storage, 'expected_cost', sln.totalcost_generation)
        table_append(storage, 'expected_fuelcost', sln.fuelcost)

#    # DEBUGGING
#    from pandas.util.testing import assert_frame_equal
#    try: 
#        gens = storage['data_generators'].copy()
#        gens.index = ['g{}'.format(g) for g in gens.index]
#        status = storage['status'].ix[1:]
#        power = storage['power'][status.columns].ix[status.index]
#        cost = storage['observed_cost'][status.columns].ix[status.index]
#        fuelcost = storage['observed_fuelcost'][status.columns].ix[status.index]
#        stcost = cost - fuelcost

#        
#        pcheck = (power * status)
#        ccheck = (cost * status)
#        stcheck = (stcost * status)
#        assert_frame_equal(pcheck, power)
#        assert_frame_equal(ccheck, cost)
#        assert_frame_equal(stcheck, stcost)
#        assert(((stcost < gens.startupcost - 1e-5) & (stcost > 1e-5) ).sum().sum() == 0)
#        
#    except AssertionError:
#        raise
        
    return storage


def load_state():
    storage = get_storage()
    user_config.update(storage['configuration'].to_dict())

    # set up the times
    startidx = int(storage['times'][0].strip('t'))
    times = TimeIndex(storage['times'].index, startidx)
    intervalhrs = (times.strings.index[1] - times.strings.index[0]
                   ).total_seconds() / 3600.0
    times._int_division = user_config.hours_commitment / intervalhrs
    times._int_overlap = user_config.hours_overlap / intervalhrs
    if len(times) <= times._int_division:
        # dont set overlap for last stage
        times._int_overlap = 0

    # create power_system
    power_system, times, scenario_tree = parse_standalone(storage, times)
    generators = power_system.generators()

    # set up initial state
    t = times.initialTime
    status = _correct_status(storage['status']).ix[t]
    for gen in generators:
        g = str(gen)
        gen.set_initial_condition(t,
                                  power=storage['power'][g][t],
                                  status=status[g],
                                  hoursinstatus=storage['hrsinstatus'][g][t])

    return power_system, times, scenario_tree


def repack_storage():
    '''compress that ballooning storage'''
    # http://stackoverflow.com/questions/13089359/mystery-when-storing-a
    # -dataframe-containing-strings-in-hdf-with-pandas
    os.system('ptrepack {f} {f}.copy; mv {f}.copy {f};'.format(
        f=user_config.store_filename))


def _add_tbl_val(storage, tablename, index, value):
    tbl = storage[tablename]
    tbl[index] = value
    storage[tablename] = tbl


def table_append(store, name, newvals):
    store[name] = store[name].append(newvals)
    return



'''Test the all of the solver links'''

from minpower import optimization, config
from test_utils import *
from coopr import pyomo

mem_tracking=False
if mem_tracking:
    from pympler.classtracker import ClassTracker
    from pympler.classtracker_stats import HtmlStats
    import objgraph
    tracker=ClassTracker()
    for cls in [pyomo.ConcreteModel,pyomo.Var,pyomo.base.var._VarElement,pyomo.Constraint,optimization.OptimizationProblem]:
        tracker.track_class(cls)

def simple_problem():
    prob=optimization.OptimizationProblem()
    if mem_tracking: tracker.create_snapshot('prob. init')
    prob.add_variable('x',low=0,high=3)
    prob.add_variable('y',low=0,high=1)
    x=prob.get_component('x')
    y=prob.get_component('y')
    prob.add_objective(y-4*x)
    prob.add_constraint('ineq',x+y<=2)
    if mem_tracking: tracker.create_snapshot('prob. created')
    return prob 


def run_one_solver(solver_name):
    prob=simple_problem()
    orig_solver = config.user_config.solver
    config.user_config.duals = False
    config.user_config.solver = solver_name 
    
    prob.solve()
    if mem_tracking: tracker.create_snapshot('prob. solved')
    status=prob.solved
    del prob
    if mem_tracking: tracker.create_snapshot('prob. deleted')
    if solver_name=='glpk' and mem_tracking:
        tracker.stats.print_summary()
        #HtmlStats(tracker=tracker).create_html('profile-simple-problem.html')
        models_left=objgraph.by_type('ConcreteModel')
        print models_left
        objgraph.show_backrefs(models_left,filename='backrefs-simple-problem.png')
    
    config.user_config.duals = True
    config.user_config.solver = orig_solver
    return status

@istest
def cplex():
    '''Test each available solver on a simple problem'''
    if 'cplex' in config.available_solvers:
        assert run_one_solver('cplex')


@istest
def glpk():
    '''Test the glpk solver on a simple problem'''
    if 'glpk' in config.available_solvers:
        assert run_one_solver('glpk')


@istest
def gurobi():
    '''Test the gurobi solver on a simple problem'''
    if 'gurobi' in config.available_solvers:
        assert run_one_solver('gurobi')        



import pandas as pd
import logging
from config import user_config
from commonscripts import update_attributes, bool_to_int, set_trace

from optimization import value, OptimizationObject
from schedule import is_init
import bidding


class Generator(OptimizationObject):
    """
    A generator model.

    :param pmin: minimum real power
    :param pmax: maximum real power
    :param minuptime: min. time after commitment in on status (hours)
    :param mindowntime: min. time after de-commitment in off status (hours)
    :param rampratemax: max. positive change in real power over 1hr (MW/hr)
    :param rampratemin: max. negative change in real power over 1hr (MW/hr)
    :param startupramplimit: max. positive change in real power over the
        first hour after startup (MW/hr)
    :param shutdownramplimit: max. negative change in real power over the
        last hour before shutdown (MW/hr)
    :param costcurveequation: text describing a polynomial cost curve ($/MWh)
      see :meth:`~bidding.parsePolynomial` for more.
    :param heatratestring: text describing a polynomial heat rate curve (MBTU/MW).
        converts to cost curve when multiplied by fuelcost.
    :param fuelcost: cost of fuel ($/MBTU)
    :param startupcost: cost to commit ($)
    :param shutdowncost: cost to de-commit ($)
    :param mustrun: flag that forces commimtent to be on

    :param name: name of the generator
    :param index: numbering of the generator
    :param bus: bus name that the generator is connected to
    """

    def __init__(self, kind='generic',
                 pmin=0, pmax=500,
                 minuptime=0, mindowntime=0,
                 rampratemax=None, rampratemin=None,
                 costcurveequation='20P',
                 heatrateequation=None, fuelcost=1,
                 bid_points=None,
                 noloadcost=0,
                 startupcost=0, shutdowncost=0,
                 startupramplimit=None,
                 shutdownramplimit=None,
                 faststart=False,
                 mustrun=False,
                 name='', index=None, bus=None):

        update_attributes(self, locals())  # load in inputs
       
        # The formulation below requires that startup ramp limits are set.
        # The defaults are the normal ramp rate limits,
        # unless Pmin so high and Rmin/max small that startup is infeasible.
        # In that case the default is Pmin.
        
        if (self.startupramplimit is None) and (self.rampratemax is not None):
            if self.pmin > self.rampratemax:
                self.startupramplimit = self.pmin
            else:
                self.startupramplimit = self.rampratemax
            
        if (self.shutdownramplimit is None) and (self.rampratemin is not None):
            if (-1 * self.pmin < self.rampratemin):
                self.shutdownramplimit = -1 * self.pmin
            else:
                self.shutdownramplimit = self.rampratemin
    
        self.fuelcost = float(fuelcost)
        
        self.is_controllable = True
        self.is_stochastic = False
        self.commitment_problem = True
        self.build_cost_model()
        self.init_optimization()

    def power(self, time=None, scenario=None):
        '''real power output at time'''
        if time is not None and is_init(time):
            return self.initial_power
        else:
            return self.get_variable('power', time, scenario=scenario, indexed=True)

    def power_available(self, time=None, scenario=None):
        '''power availble (constrained by pmax, ramprate, ...) at time'''
        if time is not None and is_init(time):
            return self.initial_power

        var_name = 'power_available' if self.commitment_problem \
            and self.reserve_required else 'power'
        return self.get_variable(var_name, time, scenario=scenario, indexed=True)

    def status(self, time=None, scenario=None):
        '''on/off status at time'''
        if self.commitment_problem or user_config.dispatch_decommit_allowed:
            if time is not None and is_init(time):
                return self.initial_status
            else:
                return self.get_variable('status', time, scenario=scenario, indexed=True)
        else:
            return 1

    def status_change(self, t, times):
        '''is the unit changing status between t and t-1'''
        if t > 0:
            previous_status = self.status(times[t - 1])
        else:
            previous_status = self.initial_status
        return self.status(times[t]) - previous_status

    def power_change(self, t, times):
        '''change in output between power between t and t-1'''
        if t > 0:
            previous_power = self.power(times[t - 1])
        else:
            previous_power = self.initial_power
        return self.power(times[t]) - previous_power

    def cost(self, time, scenario=None, evaluate=False):
        '''total cost at time (operating + startup + shutdown)'''
        return self.operatingcost(time, scenario, evaluate) + \
            self.cost_startup(time, scenario, evaluate) + \
            self.cost_shutdown(time, scenario, evaluate)

    def cost_startup(self, time, scenario=None, evaluate=False):
        if self.startupcost == 0 or not self.commitment_problem:
            return 0
        else:
            c = self.get_variable(
                'startupcost', time, scenario=scenario, indexed=True)
            return c if not evaluate else value(c)

    def cost_shutdown(self, time, scenario=None, evaluate=False):
        if self.shutdowncost == 0 or not self.commitment_problem:
            return 0
        else:
            c = self.get_variable(
                'shutdowncost', time, scenario=scenario, indexed=True)
            return c if not evaluate else value(c)

    def operatingcost(self, time=None, scenario=None, evaluate=False):
        '''cost of real power production at time (based on bid model approximation).'''
        return self.bids.output(time, scenario=scenario, evaluate=evaluate)

    def truecost(self, time, scenario=None):
        '''exact cost of real power production at time (based on exact bid polynomial).'''
        return value(self.status(time, scenario)) * self.bids.output_true(self.power(time, scenario))

    def incrementalcost(self, time, scenario=None):
        '''change in cost with change in power at time (based on exact bid polynomial).'''
        return self.bids.output_incremental(self.power(time, scenario)) if value(self.status(time, scenario)) else None

    def cost_first_stage(self, times):
        return sum(self.cost_startup(time) + self.cost_shutdown(time) for time in times)

    def cost_second_stage(self, times):
        return sum(self.operatingcost(time) for time in times)

    def getstatus(self, tend, times, status):
        return dict(
            status=value(self.status(tend)),
            power=value(self.power(tend)),
            hoursinstatus=self.gethrsinstatus(times, status))

    def gethrsinstatus(self, times, stat):
        if not self.is_controllable:
            return 0

        end_status = stat.ix[stat.index[-1]]

        if (stat == end_status).all():
            intervals = len(stat)
            hrs = intervals * times.intervalhrs
            if self.initial_status == end_status:
                hrs += self.initial_status_hours
        else:
            noneq = stat[stat != end_status]
            if len(noneq) == 0:
                intervals = 0
            else:
                intervals = len(stat.ix[noneq.index[-1]:]) - 1

            hrs = intervals * times.intervalhrs

        return hrs

    def set_initial_condition(self, time=None,
                              power=None, status=True, hoursinstatus=100):
        if power is None:
            # set default power as mean output
            power = (self.pmax - self.pmin) / 2
        if pd.isnull(power):
            raise ValueError('inital power cannot be null')
        self.initial_status = bool_to_int(status)
        self.initial_power = float(power * self.initial_status)  # note: this eliminates ambiguity of off status with power non-zero output
        self.initial_status_hours = hoursinstatus

    def build_cost_model(self):
        '''
        parse the coefficients for the polynomial bid curve
        or custom bid points definition
        '''
        bid_params = dict(
            owner=self,
            input_variable=self.power,
            min_input=self.pmin,
            max_input=self.pmax,
            status_variable=self.status
        )

        if self.bid_points is None:
            # polynomial specification
            self.cost_breakpoints = user_config.breakpoints
            if getattr(self, 'heatrateequation', None):
                self.cost_coeffs = [self.fuelcost * coef
                                    for coef in bidding.parse_polynomial(self.heatrateequation)]
            else:
                self.cost_coeffs = bidding.parse_polynomial(
                    self.costcurveequation)

            bid_params['polynomial'] = self.cost_coeffs
            bid_params['constant_term'] = self.cost_coeffs[0]
            bid_params['num_breakpoints'] = self.cost_breakpoints
            if self.noloadcost != 0:
                raise ValueError('no load cost should be defined as part of the polynomial.')

        else:
            # do some simple validation and delay construction to bidding
            # object
            min_power_bid = self.bid_points.power.min()
            max_power_bid = self.bid_points.power.max()
            if min_power_bid > self.pmin:
                self.pmin = min_power_bid
                logging.warning('{g} should have a min. power bid ({mpb}) <= to its min. power limit ({mpl})'.format(g=str(self), mpb=min_power_bid, mpl=self.pmin))

            if max_power_bid < self.pmax:
                self.pmax = max_power_bid
                logging.warning('{g} should have a max. power bid ({mpb}) >= to its max. power limit ({mpl})'.format(g=str(self), mpb=max_power_bid, mpl=self.pmax))

            bid_params['polynomial'] = None
            bid_params['bid_points'] = self.bid_points
            bid_params['constant_term'] = self.noloadcost

        self.bid_params = bid_params

    def create_variables(self, times):
        '''
        Create the optimization variables for a generator over all times.
        Also create the :class:`bidding.Bid` objects and their variables.
        '''
        self.commitment_problem = len(times) > 1
        self.add_variable('power', index=times.set, low=0, high=self.pmax)

        if self.commitment_problem or user_config.dispatch_decommit_allowed:
            self.add_variable('status', index=times.set, kind='Binary',
                              fixed_value=1 if self.mustrun else None)

        if self.commitment_problem:
            # power_available exists for easier reserve requirement
            self.reserve_required = self._parent_problem().reserve_required
            if self.reserve_required:
                self.add_variable(
                    'power_available', index=times.set, low=0, high=self.pmax)
            if self.startupcost > 0:
                self.add_variable('startupcost', index=times.set,
                                  low=0, high=self.startupcost)
            if self.shutdowncost > 0:
                self.add_variable('shutdowncost', index=times.set,
                                  low=0, high=self.shutdowncost)

        self.bids = bidding.Bid(times=times, **self.bid_params)
        return

    def create_objective(self, times):
        return sum(self.cost(time) for time in times)

    def create_constraints(self, times):
        '''create the optimization constraints for a generator over all times'''
        def roundoff(n):
            m = int(n)
            if n != m:
                raise ValueError('min up/down times must be integer number of intervals, not {}'.format(n))
            return m

        if self.commitment_problem:
            # set initial and final time constraints
            tInitial = times.initialTimestr
            tEnd = len(times)
            if self.minuptime > 0:
                up_intervals_remaining = roundoff((self.minuptime - self.initial_status_hours) / times.intervalhrs)
                min_up_intervals_remaining_init = int(
                    min(tEnd, up_intervals_remaining * self.initial_status))
            else:
                min_up_intervals_remaining_init = 0
            if self.mindowntime > 0:
                down_intervals_remaining = roundoff((self.mindowntime - self.initial_status_hours) / times.intervalhrs)
                min_down_intervals_remaining_init = int(min(tEnd, down_intervals_remaining * (self.initial_status == 0)))
            else:
                min_down_intervals_remaining_init = 0
            # initial up down time
            if min_up_intervals_remaining_init > 0:
                self.add_constraint('minuptime', tInitial, 0 >= sum([(1 - self.status(times[t])) for t in range(min_up_intervals_remaining_init)]))
            if min_down_intervals_remaining_init > 0:
                self.add_constraint('mindowntime', tInitial, 0 == sum([self.status(times[t]) for t in range(min_down_intervals_remaining_init)]))

            # initial ramp rate
            if self.rampratemax is not None:
                if self.initial_power + self.rampratemax < self.pmax:
                    E = self.power(
                        times[0]) - self.initial_power <= self.rampratemax
                    self.add_constraint('ramp lim high', tInitial, E)

            if self.rampratemin is not None:
                if self.initial_power + self.rampratemin > self.pmin:
                    E = self.rampratemin <= self.power(
                        times[0]) - self.initial_power
                    self.add_constraint('ramp lim low', tInitial, E)

            # calculate up down intervals
            min_up_intervals = roundoff(self.minuptime / times.intervalhrs)
            min_down_intervals = roundoff(self.mindowntime / times.intervalhrs)

            # reserve            
            if self.reserve_required:
                def reserve_req(model, t):
                    return self.power(t) <= self.power_available(t)
                self.add_constraint_set('max gen power avail', times.set, reserve_req)

            # ramping power
            if self.rampratemax is not None:
                def ramp_max(model, t):
                    tPrev = get_tPrev(t, model, times)
                    ramp_limit = self.rampratemax * self.status(tPrev)
                    if self.startupramplimit is not None:
                        ramp_limit += self.startupramplimit * (
                            self.status(t) - self.status(tPrev))
                        # + self.pmax * (1 - self.status(times[t]))
                    return self.power_available(t) - self.power(tPrev) <= ramp_limit

                self.add_constraint_set('ramp limit high', times.set, ramp_max)

#               # EQ19 from Carrion and Arroyo - has a conflicting
#               # definition of shutdown power, available after shutdown hour?
#                if self.shutdownramplimit is not None:
#                    # if shutting down, available power must be reduced
#                    self.add_constraint('ramp lim high shutdown', time,
#                        self.power_available(times[t-1]) <= self.pmax * self.status(time) + \
#                        self.shutdownramplimit * -1 * self.status_change(t, times)
#                        )


            if self.rampratemin is not None:
                def ramp_min(model, t):
                    tPrev = get_tPrev(t, model, times)
                    ramp_limit = self.rampratemin * self.status(t)
                    # + self.pmax * (1 - self.status(times[t-1]))
                    if self.shutdownramplimit is not None:
                        ramp_limit += self.shutdownramplimit * \
                            (-1 * (self.status(t) - self.status(tPrev)))

                    return ramp_limit <= self.power_available(t) - self.power(tPrev)
                self.add_constraint_set('ramp limit low', times.set, ramp_min)

            # start up and shut down costs
            if self.startupcost > 0:
                def startupcostmin(model, t):
                    tPrev = get_tPrev(t, model, times)
                    return self.cost_startup(t) >= self.startupcost * (
                        self.status(t) - self.status(tPrev))
                self.add_constraint_set('startup cost min', times.set, startupcostmin)

                # these tightening constraints make stochastic problems take a very long time
#                def startupcostmax(model, t):
#                    return self.cost_startup(t) <= self.startupcost * self.status(t)
#                self.add_constraint_set('startup cost max', times.set, startupcostmax)
#                
#                def startupcostmax_prev(model, t):
#                    tPrev = get_tPrev(t, model, times)
#                    return self.cost_startup(t) <= self.startupcost * (1 - self.status(tPrev))
#                self.add_constraint_set('startup cost max prev', times.set, startupcostmax_prev)
                                    
            if self.shutdowncost > 0:
                def shutdowncost(model, t):
                    tPrev = get_tPrev(t, model, times)
                    return self.cost_shutdown(t) >= self.shutdowncost * -1 * (self.status(t) - self.status(tPrev))
                self.add_constraint_set('shutdown cost', times.set, shutdowncost)

            # note: costs must be >= constraints
            # for very large problems with >= constraints,
            # the solver may not always force the costs down to their minimum
            # however, the == formulation is incorrect and will not allow
            # unit shutdowns if the unit has a startup cost
            # solution is to use the min and max constraints together.

            # TODO: convert these to constraint list form
            for t, time in enumerate(times):

                # min up time
                if t >= min_up_intervals_remaining_init and self.minuptime > 0:
                    no_shut_down = range(t, min(tEnd, t + min_up_intervals))
                    min_up_intervals_remaining = min(tEnd - t, min_up_intervals)
                    E = sum([self.status(times[s]) for s in no_shut_down]) >= min_up_intervals_remaining * self.status_change(t, times)
                    self.add_constraint('min up time', time, E)
                # min down time
                if t >= min_down_intervals_remaining_init and self.mindowntime > 0:
                    no_start_up = range(t, min(tEnd, t + min_down_intervals))
                    min_down_intervals_remaining = min(
                        tEnd - t, min_down_intervals)
                    E = sum([1 - self.status(times[s]) for s in no_start_up]) >= min_down_intervals_remaining * -1 * self.status_change(t, times)
                    self.add_constraint('min down time', time, E)
                


        # min/max power limits
        # these always apply (even if not a UC problem)
        if self.pmin > 0:
            def min_power(model, t):
                return self.power(t) >= self.status(t) * self.pmin
            self.add_constraint_set('min gen power', times.set, min_power)

        def max_power(model, t):
            return self.power_available(t) <= self.status(t) * self.pmax
        self.add_constraint_set('max gen power', times.set, max_power)

        return

    def __str__(self):
        return 'g{ind}'.format(ind=self.index)

    def __int__(self):
        return self.index


def get_tPrev(t, model, times):
    return model.times.prev(t) if t != model.times.first() else times.initialTime


class Generator_nonControllable(Generator):
    """
    A generator with a fixed schedule.
    """
    def __init__(self,
                 schedule=None,
                 fuelcost=1,
                 costcurveequation='0',
                 bid_points=None,
                 noloadcost=0,
                 mustrun=False,
                 faststart=False,
                 sheddingallowed=False,
                 pmin=0, pmax=None,
                 name='', index=None, bus=None, kind='wind',
                 observed_values=None,
                 **kwargs):

        update_attributes(self, locals())  # load in inputs
        self.is_controllable = False
        self.startupcost = 0
        self.shutdowncost = 0
        self.fuelcost = float(fuelcost)
        self.build_cost_model()
        self.init_optimization()
        self.is_stochastic = False
        self.shedding_mode = sheddingallowed and user_config.economic_wind_shed

    def power(self, time, scenario=None):
        if self.shedding_mode:
            power = self.get_variable('power_used', time,
                                      scenario=scenario, indexed=True)
        else:
            power = self.power_available(time)
        return power

    def status(self, time=None, scenarios=None):
        return True

    def power_available(self, time=None, scenario=None):
        return self.get_parameter('power', time, indexed=True)

    def shed(self, time, scenario=None, evaluate=False):
        Pused = self.power(time, scenario=scenario)
        Pavail = self.power_available(time, scenario=scenario)
        if evaluate:
            Pused = value(Pused)
            Pavail = value(Pavail)
        return Pavail - Pused

    def set_initial_condition(self, time=None,
                              power=None, status=None, hoursinstatus=None):
        self.initial_power = 0
        self.initial_status = 1
        self.initial_status_hours = 0

    def getstatus(self, tend, times=None, status=None):
        return dict(
            status=1,
            power=self.power(tend),
            hoursinstatus=0)

    def create_variables(self, times):
        if self.shedding_mode:
            self.create_variables_shedding(times)
        self.add_parameter('power', index=times.set,
                           values=dict([(t, self.get_scheduled_ouput(t)) for t in times]))
        self.create_bids(times)

    def create_bids(self, times):
        self.bids = bidding.Bid(
            polynomial=self.cost_coeffs,
            owner=self,
            times=times,
            fixed_input=True,
            status_variable=lambda *args: True,
            input_variable=self.power
        )

    def create_variables_shedding(self, times):
        self.add_variable('power_used', index=times.set, low=0)

    def create_constraints(self, times):
        if self.shedding_mode:
            for time in times:
                self.add_constraint('max_power', time,
                                    self.power(time) <= self.power_available(time))

    def cost(self, time, scenario=None, evaluate=False):
        return self.operatingcost(time, scenario=scenario, evaluate=evaluate)

    def operatingcost(self, time=None, scenario=None, evaluate=False):
        return self.bids.output(time, scenario=scenario) + \
            user_config.cost_wind_shedding * self.shed(time, scenario=scenario, evaluate=evaluate)

    def truecost(self, time, scenario=None):
        return self.cost(time)

    def incrementalcost(self, time, scenario=None):
        return self.bids.output_incremental(self.power(time))

    def cost_startup(self, time, scenario=None):
        return 0

    def cost_shutdown(self, time, scenario=None):
        return 0

    def cost_first_stage(self, times):
        return 0

    def cost_second_stage(self, times):
        return sum(self.cost(time) for time in times)

    def get_scheduled_ouput(self, time):
        return float(self.schedule.ix[time])

    def set_power_to_observed(self, times):
        power = self.power_available()
        for time in times:
            power[time] = self.observed_values[time]


class Generator_Stochastic(Generator_nonControllable):
    """
    A generator with a stochastic power output.
    """
    def __init__(self,
                 scenario_values=None,
                 costcurveequation='0',
                 fuelcost=1,
                 bid_points=None, noloadcost=0,
                 mustrun=False,
                 faststart=False,
                 pmin=0, pmax=None,
                 name='', index=None, bus=None, kind='wind',
                 observed_values=None,
                 schedule=None,
                 sheddingallowed=False,
                 **kwargs):
        update_attributes(self, locals())  # load in inputs
        self.is_controllable = False

        self.is_stochastic = not \
            (user_config.perfect_solve or user_config.deterministic_solve)
        self.build_cost_model()
        self.init_optimization()
        self.startupcost = 0
        self.shutdowncost = 0
        self.fuelcost = float(fuelcost)
        self.shedding_mode = sheddingallowed and user_config.economic_wind_shed

    def power(self, time, scenario=None):
        return self.get_variable(
            'power_used' if self.shedding_mode else 'power',
            time=time, scenario=scenario, indexed=True)

    def power_available(self, time=None, scenario=None):
        return self.get_variable('power',
                                 time=time, scenario=scenario, indexed=True)

    def _get_scenario_values(self, times, s=0):
        # scenario values are structured as a pd.Panel
        # with axes: day, scenario, {prob, [hours]}
        return self.scenario_values[times.Start][
            range(len(times))].ix[s].values.tolist()

    def create_variables(self, times):
        if self.shedding_mode:
            self.create_variables_shedding(times)
        
        if self.is_stochastic:
            # initialize parameter set to first scenario value
            scenario_one = self._get_scenario_values(times, s=0)
            self.add_parameter('power', index=times.set, nochecking=True,
                values=dict(zip(times, scenario_one)))
        else:
            # set to forecast values
            self.add_parameter('power', index=times.set,
               values=dict([(t, self.get_scheduled_ouput(t)) for t in times]))

        self.create_bids(times)
        return



from minpower.config import user_config, parse_command_line_config
from minpower.get_data import _load_raw_data, _parse_raw_data, setup_times
from minpower.commonscripts import joindir, set_trace
from minpower.powersystems import PowerSystem
from minpower.solve import create_solve_problem
import pandas as pd
import argparse 


def initial_dispatch(directory='.', output_filename='initial.csv'):
    user_config.directory = directory
    user_config.dispatch_decommit_allowed = True
    
    generators_data, loads_data, lines_data, init_data = _load_raw_data()
    
    init_data = pd.DataFrame()
    
    timeseries, times, generators_data, loads_data = \
        setup_times(generators_data, loads_data)

    # get rid of a bunch of stuff to make this an ED problem    
    generators_data['power'] = None
    loads_data['power'] = None
    for i, snm in generators_data.pop('schedulename').dropna().iterkv():
        generators_data.ix[i, 'power'] = timeseries[snm].values[0]

    for i, snm in loads_data.pop('schedulename').dropna().iterkv():
        loads_data.ix[i, 'power'] = timeseries[snm].values[0]
                
    for i, snm in generators_data.pop('observedname').dropna().iterkv():
        generators_data.ix[i, 'power'] = timeseries[snm].values[0]
    
    valid_columns = pd.Index([
        'name', 'pmin', 'pmax',
        'fuelcost', 'heatrateequation', 'costcurveequation',
        'sheddingallowed',
        'power'])
    
    generators_data = generators_data[generators_data.columns & valid_columns]
        
    generators, loads, lines, times, scenario_values, data = \
        _parse_raw_data(generators_data, loads_data, 
            lines_data, init_data)

    power_system = PowerSystem(generators, loads, lines)
    sln = create_solve_problem(power_system, times)

    # save a csv of the dispatch
    dispatch = pd.DataFrame({
        'power': sln.generators_power.ix[0].values, 
        'status': sln.generators_status.ix[0].values},
        index=[gen.name for gen in sln.generators])
    dispatch.ix[dispatch.status==0, 'power'] = 0
    dispatch.index.name = 'name'
    dispatch.to_csv(joindir(user_config.directory, output_filename))
    
    print(dispatch)

def main():
    parser = argparse.ArgumentParser('Initial dispatch tool',
        description='''Get the initial dispatch for a case. Runs a economic
        dispatch for the first hour of a commitment problem, but allows 
        generators to turn off to minimize cost.''')
    parser.add_argument('--output_filename', default='initial.csv')

    # parse all of the standard minpower options
    args = parse_command_line_config(parser)
    initial_dispatch(args['directory'], args['output_filename'])

if __name__ == '__main__':
    main()



# minpower's setup script

# make a new release by using 
# seed release --dry-run
# and then 
# seed release

import sys
from setuptools import setup, find_packages

version_number = open('minpower/__init__.py').read().split('"')[1].rstrip('"')

python_version = sys.version_info

install_requires=[
    'Coopr>=3.3.7114',
    'numpy>=1.6.1',
    'pandas>=0.10',
    ]
    
if python_version[0] == 2 and python_version[1] < 7:
    install_requires.append('ordereddict>=1.1')
    
    # syntax changes currently result in failures on python 2.6
    # https://github.com/adamgreenhall/minpower/issues/8
    # https://travis-ci.org/adamgreenhall/minpower/jobs/4433662
    sys.exit('minpower currently requires Python 2.7')
    

setup(
    name = "minpower",
    version = version_number,
    download_url = "https://github.com/adamgreenhall/minpower" + \
        "/zipball/v{v}".format(v=version_number),
    entry_points="""
    [console_scripts]
    minpower = minpower.solve:main
    standalone_minpower = minpower.solve:standaloneUC
    scheduler_minpower = minpower.experiments.scheduler_minpower:main
    initial_dispatch = minpower.experiments.get_initial_dispatch:main
    """,

    package_data={
        'minpower.configuration': ['minpower.cfg'],
        'minpower.tests': [
            '*.csv', 
            '*/*.csv',
            '*/*/*.csv']},

    install_requires=install_requires,
    
    tests_require=[
        'nose',
        'coverage',
        'objgraph'
        ],
    
    # it helps to have seed if you are going to make releases
    # but it is not required for setup

    description = "power systems optimization made beautiful",
    author = "Adam Greenhall",
    author_email = "minpower@adamgreenhall.com",
    url = "http://minpowertoolkit.com/",

    packages = find_packages(),
    keywords = ["power systems","optimization"],

    classifiers = [
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        ],
    long_description = """
power systems tools made beautiful
-----------------------------------------

* Solves ED, OPF, and UC problems.
* Problems can be defined in simple spreadsheets.
* Visualizations are created for the answers.
* Many solvers are supported.


* `Full documentation and tutorials <http://minpowertoolkit.com>`_
* `Actively developed <http://github.com/adamgreenhall/minpower>`_

"""
)



"""
Time and schedule related models.
"""

from datetime import timedelta, datetime
from commonscripts import *
from operator import attrgetter


def get_schedule(filename):
    return ts_from_csv(filename)


def make_times_basic(N):
    '''make a :class:`schedule.TimeIndex` of N times with hourly interval'''
    return TimeIndex(date_range('00:00:00', periods=N, freq='H'))


def just_one_time():
    '''make a TimeIndex with just one time in it'''
    return make_times_basic(1)


def make_constant_schedule(times, power=0):
    return Series(power, times.strings.values)


class TimeIndex(object):
    '''a list of times (underlying model is pandas.Index)'''
    def __init__(self, index, str_start=0):
        strings = ['t%02d' % (i + str_start) for i in range(len(index))]
        self.times = index.copy()
        self.strings = Series(strings, index=self.times)
        self._set = self.strings.values.tolist()

        self.get_interval()
        self.Start = self.times[0]
        self.startdate = self.Start.date()

        self.End = self.times[-1] + self.interval
        self.span = self.End - self.Start
        self.spanhrs = hours(self.span)

        self.set_initial()

        self._int_overlap = 0
        self._int_division = len(self)
        self._str_start = str_start

    def set_initial(self, initialTime=None):
        if initialTime:
            self.initialTime = initialTime
        else:
            self.initialTime = pd.Timestamp(self.Start - self.interval)
            self.initialTime.index = 'Init'
        self.initialTimestr = 'tInit'

    def get_interval(self):
        freq = self.strings.index.freq
        if freq is not None:
            self.interval = freq
            if self.interval.freqstr == 'H':
                self.intervalhrs = self.interval.n
            else:
                self.intervalhrs = self.interval.nanos / 1.0e9 / 3600.0
        else:
            self.interval = self.times[1] - self.times[0]
            self.intervalhrs = self.interval.total_seconds() / 3600.0
        return

    def __contains__(self, item):
        return item in self.times

    def __repr__(self):
        return repr(self.times)

    def __len__(self):
        return len(self.times)

    def __getitem__(self, i, circular=False):
        if i == -1 and not circular:
            return self.initialTime
        else:
            return self.strings[i]

    def last(self):
        return self.__getitem__(-1, circular=True)

    def __getslice__(self, i, j):
        return self.strings[i:j]

    def non_overlap(self):
        if self._int_overlap > 0:
            return TimeIndex(self.strings.index[:-1 - self._int_overlap + 1], self._str_start)
        else:
            return self
        return

    def post_horizon(self):
        if len(self) > self._int_division + 1:
            str_start = int(self.strings.ix[self._int_division + 1].strip('t'))
            return TimeIndex(self.strings.index[self._int_division + 1:], str_start)
        else:
            return Series()
        return

    def last_non_overlap(self):
        return self.strings.index[-1 - self._int_overlap]

    def subdivide(self, division_hrs=24, overlap_hrs=0):
        int_division = int(division_hrs / self.intervalhrs)
        int_overlap = int(overlap_hrs / self.intervalhrs)
        subsets = []
        for stg in range(int(len(self) / int_division)):
            start = stg * int_division
            end_point = start + int_division + int_overlap
            end_point = min(end_point, len(self))

            subset = TimeIndex(self.times[start:end_point], start)
            subset._int_overlap = int_overlap
            subset._int_division = int_division
            subsets.append(subset)

        if len(subsets[-1]) <= int_division:
            subsets[-1]._int_overlap = 0
        return subsets


def is_init(time):
    return getattr(time, 'index', None) == 'Init'



"""
Provide the defaults and configuration for other modules.
`user_config` is treated as a global in minpower.
"""
import os
import sys
from ConfigParser import SafeConfigParser
from commonscripts import DotDict, joindir

parser = SafeConfigParser()
parser.read([
    # the minpower default set, from the minpower/configuration directory
    joindir(os.path.split(__file__)[0], 'configuration/minpower.cfg'),
    # the user's global overrides, from the home directory
    os.path.expanduser('~/minpower.cfg'),
    os.path.expanduser('~/.minpowerrc'),
])

option_types = dict(
    duals=bool,
    breakpoints=int,
    hours_commitment=int,
    hours_overlap=int,
    
    cost_load_shedding=float,
    cost_wind_shedding=float,
    economic_wind_shed=bool,
    dispatch_decommit_allowed=bool,
    solver=str,
    mipgap=float,
    solver_time_limit=float,
    
    reserve_fixed=float,
    reserve_load_fraction=float,

    faststart_resolve=bool,

    visualization=bool,
    logging_level=int,
    problem_file=bool,
    output_prefix=bool,
    debugger=bool,

    scenarios=int,
    deterministic_solve=bool,
    perfect_solve=bool,
    scenarios_directory=str,

    standalone=bool,
    pid=str,

    wind_forecast_adder=float,
    wind_multiplier=float,
    wind_capacity_factor=float,
    wind_error_multiplier=float,

    load_multiplier=float,
    load_adder=float,

    pmin_multiplier=float,
    ramp_limit_multiplier=float,

    ignore_minhours_constraints=bool,
    ignore_ramping_constraints=bool,
    ignore_pmin_constraints=bool,

    on_complete_script=str,

    file_gens=str,
    file_loads=str,
    file_lines=str,
    file_init=str,

    # HACKs to help out resetting the config in testing
    directory=str,
    store_filename=str,
)

def parser_value(parser, section, key, opt_type):
    meth = 'get'
    if opt_type==bool:
        meth += 'boolean'
    elif opt_type==str:
        pass
    else:
        meth += str(opt_type).split("'")[1]
    return getattr(parser, meth)(section, key)

def parse_config(parser):
    minpower_config = DotDict(
        dict( [(k, parser_value(parser, 'minpower', k, v)) \
            for k, v in option_types.iteritems()]))
    
    s = 'scheduler'
    scheduler_config = DotDict(dict(
        email=parser.get(s, 'email'),
        hours_limit=parser.getint(s, 'hours_limit'),
        memory=parser.getint(s, 'memory'),
        scheduler_mode=parser.get(s, 'scheduler_mode'),
        verbose=parser.getboolean(s, 'verbose'),
        ))
    
    return minpower_config, scheduler_config

user_config, scheduler_config = parse_config(parser)

available_solvers = []
for solver in parser.options('available_solvers'):
    if parser.getboolean('available_solvers', solver):
        available_solvers.append(solver)

def filter_non_defaults(x, defaults):
    return dict((k, v) for k, v in x.iteritems() if \
        (k not in defaults) or (v != defaults[k])
        )

def get_dir_config(directory):
    '''
    pull the config from a specific directory.
    this allows post-load updating of the user_config.
    '''
    dirparser = SafeConfigParser()
    dirparser.read([
        # the minpower default set, from the minpower/configuration directory
        # need this to set the defaults
        joindir(os.path.split(__file__)[0], 'configuration/minpower.cfg'),
        # need the home directory overrides too
        os.path.expanduser('~/minpower.cfg'),
        os.path.expanduser('~/.minpowerrc'),
        # the directory's defaults
        joindir(directory, 'minpower.cfg'),
    ])
    
    new_user_config, new_scheduler_conifg = parse_config(dirparser)
    # return (filter_non_defaults(new_user_config, user_config),
    #    filter_non_defaults(new_scheduler_conifg, scheduler_config))
    return new_user_config, new_scheduler_conifg
    
def parse_command_line_config(parser, preparsed_args=None):
    # get the directory first
    
    parser = setup_parser_args(parser)
    
    directory = parser.parse_args(preparsed_args).directory
        
    # an extra config parse to make sure that if you are calling a case but not 
    # in the case directory, the case config gets used
    new_user_config, new_scheduler_config = get_dir_config(directory)
    user_config.update(new_user_config)
    scheduler_config.update(new_scheduler_config)
    
    parser.set_defaults(**dict(user_config))

    #figure out the command line arguments
    clargs = vars(parser.parse_args(preparsed_args))

    user_config.update(clargs)

    user_config.update(dict(directory=directory))
    
    if clargs.pop('show_config'):
        from pprint import pprint
        pprint(user_config)
        sys.exit()

    return clargs

def add_opt(parser, name, shortcut=None, help=""):
    kwds = dict(help=help, type=option_types[name], default=user_config[name])
    callers = ['--{n}'.format(n=name)]
    
    if shortcut:
        callers.append(shortcut)
    if kwds['type'] == bool:
        kwds['action'] = 'store_true'
        kwds.pop('type')

    parser.add_argument(*callers, **kwds)
    
def setup_parser_args(parser):
    parser.add_argument('directory', type=str,
        help='the direcory of the problem you want to solve')

    add_opt(parser, 'solver', '-s', 
        help='the solver name (e.g. cplex, gurobi, glpk)')
    add_opt(parser, 'visualization', '-v',
        help='save a visualization of the solution')
        
    add_opt(parser, 'breakpoints', '-b',
        help='number of breakpoints to use in piecewise linearization of polynomial costs')
    add_opt(parser, 'hours_commitment','-c',
        help='number hours per commitment in a rolling UC (exclusive of overlap)')
    add_opt(parser, 'hours_overlap','-o',
        help='number hours to overlap commitments in a rolling UC')


    solver_opt = parser.add_argument_group('Solver options')
    add_opt(solver_opt, 'mipgap',
        help='the MIP gap solution tolerence')
    add_opt(solver_opt, 'solver_time_limit', 
        help='the MIP solver time limit (in seconds)')

        
    reserve = parser.add_argument_group('Reserve',
        'Does the system require reserve? The default is no reserve.')
    add_opt(reserve, 'reserve_fixed',
        help='the static amount of reserve required at all times (in MW)')
    add_opt(reserve, 'reserve_load_fraction',
        help='fraction of the total system load which is required as reserve')

    add_opt(parser, 'duals', '-d',
        help='flag to get the duals, or prices, of the optimization problem')
    add_opt(parser, 'dispatch_decommit_allowed',
        help='flag to allow de-commitment of units in an ED -- useful for getting initial conditions for UCs')
    
    add_opt(parser, 'cost_load_shedding', 
        help='the cost to the system to shed a MWh of load')
    add_opt(parser, 'cost_wind_shedding',
        help='the cost to the system to shed a MWh of wind energy')
    add_opt(parser, 'economic_wind_shed',
        help='is wind allowed to be shed for economic reasons ' + \
        '(default is to allow wind shedding only if infeasible)')
        
    
    stochastic = parser.add_argument_group('Stochastic UC',
        'options to modify the behavior of a stochastic problem')
    add_opt(stochastic, 'scenarios',
        help='limit the number of scenarios to N')        
    add_opt(stochastic, 'faststart_resolve', '-F',
        help="""allow faststart units which are off to be
                started up during resolve with observed wind values""")

    stochastic_mode = stochastic.add_mutually_exclusive_group()
    add_opt(stochastic_mode, 'deterministic_solve', '-D',
        help='solve a stochastic problem deterministically using the forecast_filename paramter')
    add_opt(stochastic_mode, 'perfect_solve', '-P',
        help='solve a stochastic problem with perfect information')
    add_opt(stochastic_mode, 'scenarios_directory', 
        help='override scenarios directory for stochastic problem')
        

    add_opt(parser, 'standalone', '-m', 
        help='Make each multi-day commitment its own subprocess (helps with memory issues).')
    add_opt(parser, 'output_prefix', '-p',
        help='Prefix all results files with the process id (for a record of simulataneous solves)')
    add_opt(parser, 'pid', 
        help='Expicitly specify the standalone storage process id')

    debugging = parser.add_argument_group('Debugging tools')
    add_opt(debugging, 'debugger',
        help='use pdb when an error is raised')
    add_opt(debugging, 'problem_file',
        help='flag to write the problem formulation to a problem.lp file')        
    add_opt(parser, 'logging_level', 
        help='set the level of detail for logging')
    debugging.add_argument('--profile', action="store_true", default=False,
        help='run cProfile and output to minpower.profile')
    debugging.add_argument('--show_config', action='store_true', default=False,
        help='just show the configuration and quit')
    debugging.add_argument('--keep_lp_files', action='store_true', default=False,
        help='keep a record of all of the solver lp files')

    
    ts = parser.add_argument_group('Timeseries modifiers',
        'Alter the timeseries after parsing the data.')
    
    wind_modifier = ts.add_mutually_exclusive_group()
    add_opt(wind_modifier, 'wind_multiplier', 
        help='scale the wind power by this factor')
    add_opt(wind_modifier, 'wind_capacity_factor', 
        help='scale the wind power to match the capacity factor (as a percentage of peak load power)')
    
    add_opt(ts, 'wind_error_multiplier', 
        help='scale the wind power forecast error by this factor')
    add_opt(ts, 'wind_forecast_adder', 
        help='add a fixed amount (e.g. bias) to the wind power forecast')
    add_opt(ts, 'load_multiplier', 
        help='scale the load power by this factor')
    add_opt(ts, 'load_adder', 
        help='add a fixed amount to the load power')
    
    constraints = parser.add_argument_group('Ignore/relax constraints',
        'Ignore or relax sets of constraints to allow for feasible solutions.')
    add_opt(constraints, 'ignore_minhours_constraints', 
        help='drop the min up/down time constraints on the generators')
    add_opt(constraints, 'ignore_ramping_constraints', 
        help='drop the min ramping power constraints on the generators')
    add_opt(constraints, 'ignore_pmin_constraints', 
        help='drop the min power constraints on the generators')
    add_opt(constraints, 'pmin_multiplier', 
        help='scale the generator minimum power limits by this factor')
    add_opt(constraints, 'ramp_limit_multiplier', 
        help='scale the generator ramp power limits by this factor')

    filenames = parser.add_argument_group('Filenames',
        'Change the default filenames to use when loading data.')
    add_opt(filenames, 'file_gens')
    add_opt(filenames, 'file_loads')
    add_opt(filenames, 'file_lines')
    add_opt(filenames, 'file_init')

    add_opt(parser, 'on_complete_script',  
        help='run a script on completion of the minpower script')

    # NOTE - don't let defaults creep into this defenition
    # that makes resetting the defaults during testing very hard 
    
    return parser



"""
Short, commonly used scripts.
Many of them are one liners.
"""

import os
import sys
import csv
import logging
import itertools
import operator
import datetime
from contextlib import contextmanager

import numpy as np
import pandas as pd
from pandas import DataFrame, Series, date_range

from pprint import pprint
try:  # for development
    from ipdb import set_trace  # pudb
    import IPython
    from IPython.frontend.terminal.embed import InteractiveShellEmbed
#    from IPython.config.loader import Config
#
#    cfg = Config()
#    cfg.InteractiveShellEmbed.prompt_in1="myprompt [\\#]> "
#    cfg.InteractiveShellEmbed.prompt_out="myprompt [\\#]: "
#    cfg.InteractiveShellEmbed.profile=ipythonprofile
    ipython_shell = InteractiveShellEmbed(
    )  # config=cfg, user_ns=namespace, banner2=banner)
except:
    from pdb import set_trace


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """
    def __init__(self, logger=logging, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ''

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())

@contextmanager
def not_quiet():
    yield
    
@contextmanager
def quiet():
    sys.stderr.flush()
    sys.stdout.flush()
    devnull = open(os.devnull, "w")
    sys.stdout = devnull
    sys.stderr = devnull

    logger = logging.getLogger()
    current_log_level = logger.level
    logger.setLevel(logging.CRITICAL)

    try:
        yield
    finally:
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        logger.setLevel(current_log_level)


def gen_time_dataframe(generators, times, values=()):
    kwargs = dict(columns=[str(g) for g in generators])
    try:
        kwargs['index'] = times.strings.index
    except AttributeError:
        kwargs['index'] = times

    if values:
        values = np.array(values)
        if values.shape != (len(times), len(generators)):
            values = values.T
        df = DataFrame(values, **kwargs)
    else:
        df = DataFrame(**kwargs)

    df.index.name = 'time'
    return df


def ts_from_csv(filename, index_col=0, squeeze=True, timezone=None, is_df=True, **kwargs):
    kwargs['header'] = 0 if is_df else None

    df = pd.read_csv(filename, index_col=index_col, squeeze=squeeze, **kwargs)
    df.index = pd.DatetimeIndex(df.index)
    if timezone is not None:
        # pandas seems to convert any stamps to UTC in the DatetimeIndex call
        df.index = df.index.tz_localize('UTC').tz_convert(timezone)
    return df


def bool_to_int(x):
    return 1 if x else 0


class DotDict(dict):
    '''a dict with dot notation access'''
    def __getattr__(self, attr):
        return self.get(attr)

    def copy(self):
        return DotDict(dict(self))

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

###### matrix stuff #######


def elementwiseAdd(La, Lb):
    return map(operator.add, La, Lb)


def transpose(listoflists):
    return map(None, *listoflists)


def flatten(listoflists):
    '''Flatten one level of nesting'''
    return list(itertools.chain.from_iterable(listoflists))


def within(x, val=0, eps=1e-3):
    return (val - eps) <= x <= (val + eps)


def replace_all(seq, obj, replacement):
    def with_index(seq):
        for i in xrange(len(seq)):
            yield i, seq[i]

    for i, elem in with_index(seq):
        if elem == obj:
            seq[i] = replacement
    return seq


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return itertools.izip(a, b)

##### csv stuff #####

def writeCSV(fields, data, filename):
    with open(filename, 'w+') as f:
        writer = csv.writer(f)
        writer.writerow(fields)
        writer.writerows(data)

#################### string stuff ##################


def convert_str2num(s):
    s = s.strip()
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s


def indexCaseSpaceInsensitive(L, s):
    return map(drop_case_spaces, L).index(drop_case_spaces(s))


def drop_case_spaces(s):
    '''get rid of spaces in a string and make lower case. will also work with list of strings'''
    try:
        return s.lower().replace(' ', '').replace('_', '')
    except AttributeError:
        if s is None:
            return None
        elif isinstance(s, list):
            return map(drop_case_spaces, s)


def to_percent(val, digits=0):
    return '{p:.{d}%}'.format(p=val, d=digits)


##################### file stuff ###########################
def splitFilename(fullPathFilenm):
    '''split a filename into its directory, filename, and extension'''
    (dirNm, fullFilenm) = os.path.split(fullPathFilenm)
    (fileNm, extNm) = os.path.splitext(fullFilenm)
    return dirNm, fileNm, extNm


def joindir(dir, file):
    return os.path.join(dir, file)

################### time stuff ###########################


def hours(t):
    try:
        return t.days * 24.0 + t.seconds / 3600.0  # t is a datetime object
    except AttributeError:
        return datetime.timedelta(hours=t)  # t is a number

####################### class stuff #######################


def getattrL(L, attribute='name'):
    '''get the attribute of each class instance in a list'''
    return [getattr(item, attribute) for item in L]


def update_attributes(instance, variables, exclude=['self'], include=None):
    """Update instance attributes

    For example, update(self, locals())

    instance: Instance to update via setattr()
    variables: Dictionary of variables
    exclude: Variables to explicitly exclude, defaults to ['self'].
    include: Variables to explicitly include. If set, include overides exclude.
    """
    if include is not None:
        [setattr(instance, k, v) for k, v in variables.items() if k in include]
    else:
        if 'self' not in exclude:
            exclude.append('self')
        [setattr(
            instance, k, v) for k, v in variables.items() if k not in exclude]


def debug_frame_unequal(left, right, tol=1e-5):
    diff = left - right
    different = diff[ (diff < -tol) | (diff > tol) ]
    cols = different.columns[different.any()]
    idx = different.ix[different.any(axis=1)].index
    
    print(left[cols].ix[idx])
    print(right[cols].ix[idx])
    return cols, idx            



"""
A selection of integration tests. These are designed to find failures that
the unit tests don't pick up.
"""
import os
import pandas as pd
import nose
from nose.tools import istest
from pandas.util.testing import assert_frame_equal, assert_series_equal

from minpower.solve import solve_problem as solve_dir
from minpower.config import user_config
from test_utils import reset_config, with_setup


def basic_checks(sln):
    cost = sln.totalcost_generation
    power = sln.generators_power.ix[cost.index]
    status = sln.generators_status.ix[cost.index]

    # make sure that if status = 0 then power also = 0
    Pcheck = power * status
    assert_frame_equal(Pcheck, power)

    # make sure that if status = 0 then cost also = 0
    Ccheck = cost * status
    assert_frame_equal(Ccheck, cost)

    # make sure that generation meets load (plus any shed)
    shed = sln.load_shed_timeseries
    try: scheduled = sln.power_system.total_scheduled_load()
    except AttributeError: 
        # look up the scheduled load from the timeseries data
        loads = [col for col in \
            sln.store['data_timeseries'].columns if col.startswith('d')]
        scheduled = sln.store['data_timeseries'][loads].sum(axis=1)
    try: scheduled.index = shed.index
    except AssertionError:
        # not the same length due to overlap
        scheduled = scheduled.ix[scheduled.index[:len(shed)]]
        scheduled.index = shed.index

    assert_series_equal(scheduled - shed, power.sum(axis=1))

#    except:
#        wind_gen = sln.power_system.get_generator_with_observed()
#        gen_scheduled = pd.DataFrame(dict(
#            wind_obs=wind_gen.observed_values,
#            other_sched=sum(gen.schedule \
#                for gen in sln.power_system.generators() \
#                if not gen.is_controllable and gen != wind_gen)
#            )).ix[:len(shed.index)]  
#        gen_scheduled.index = shed.index
#        
#        print('controllable_requirement')
#        print(scheduled - gen_scheduled.sum(axis=1))
#        raise


this_directory = os.path.dirname(__file__)
default_config = user_config.copy()

def run_case(name, basedir=this_directory, **kwds):
    '''
    a programatic way to do (nearly) the same thing
    as calling the minpower script
    (nearly: because prepends basedir to the directory name you are solving)
    '''
    # reset user_config to defaults first
    user_config.update(default_config)
    assert(user_config == default_config)
    user_config.breakpoints = 3 # for speedier testing
    user_config.update(kwds)
    sln = solve_dir(os.path.join(basedir, name))
    basic_checks(sln)
    return sln


@istest
def run_uc():
    run_case('uc')

@istest
def run_uc_rolling():
    run_case('uc-rolling')

@istest
def run_ed():
    run_case('ed')

@istest
def run_opf():
    run_case('opf')
    

# just test that the visulizations all work
# if matplotlib is installed
@istest
@with_setup(teardown=reset_config)
def ed_visualization():
    try:
        import matplotlib
    except ImportError:
        raise nose.SkipTest('visualizations require matplotlib')

    run_case('ed', visualization=True)


@istest
@with_setup(teardown=reset_config)
def opf_visualization():
    try:
        import matplotlib
    except ImportError:
        raise nose.SkipTest('visualizations require matplotlib')

    run_case('opf', visualization=True)


@istest
@with_setup(teardown=reset_config)
def uc_visualization():
    try:
        import matplotlib
    except ImportError:
        raise nose.SkipTest('visualizations require matplotlib')

    duals = user_config.solver != 'glpk'

    run_case('uc-rolling', visualization=True, duals=duals)
